#! /usr/bin/env python

from __future__ import absolute_import
import fnmatch, logging, logtool, requests, retryp
from addict import Dict
from .flattendict import flatten_dict

LOG = logging.getLogger (__name__)
logging.basicConfig (level = logging.WARN)

@retryp.retryp (expose_last_exc = True)
@logtool.log_call
def get_data (mq_conn, endpoint):
  url = "{url}{endpoint}".format (endpoint = endpoint, **mq_conn)
  rc = requests.get (url, auth = (mq_conn["user"], mq_conn["passwd"]))
  if isinstance (rc.json (), list):
    return [Dict (flatten_dict (i)) for i in rc.json ()]
  return Dict (flatten_dict (rc.json ()))

@logtool.log_call
def message_stats (name, prefix, data, stats):
  stubs = [
    "ack",
    "ack_details.rate",
    "deliver",
    "deliver_details.rate",
    "deliver_get",
    "deliver_get_details.rate",
    "deliver_no_ack",
    "deliver_no_ack_details.rate",
    "get_no_ack",
    "get_no_ack_details.rate",
    "publish",
    "publish_details.rate",
    "redeliver",
    "redeliver_details.rate",
  ]
  keys = ["%s.%s" % (prefix, k) for k in stubs]
  aggregate (name, keys, data, stats)

@logtool.log_call
def aggregate (name, keys, data, stats):
  for k in keys:
    key = "{}.{}".format (name, k)
    value = data[k]
    if value == {}:
      continue
    if value in (False, True):
      value = 1 if value else 0
    stats[key] = value

@logtool.log_call
def pattern_match (key, patterns):
  for pattern in patterns:
    p = pattern.strip ()
    if fnmatch.fnmatch (key, p): # Discard matches
      return True
  return False

@logtool.log_call
def mqreport (mp_conn, ignore_queues = None):
  stats = Dict ()
  mp_conn = Dict (mp_conn)
  data = get_data (mp_conn, "api/nodes")
  keys = [
    "disk_free",
    "disk_free_alarm",
    "fd_total",
    "fd_used",
    "mem_alarm",
    "mem_used",
    "proc_total",
    "proc_used",
    "run_queue",
    "running",
    "sockets_total",
    "sockets_used",
  ]
  for d in data:
    name = "nodes." + d.name.split ("@")[1]
    aggregate (name, keys, d, stats)
  data = get_data (mp_conn, "api/overview")
  keys = [
    "object_totals.channels",
    "object_totals.connections",
    "object_totals.consumers",
    "object_totals.exchanges",
    "object_totals.queues",
    "queue_totals.messages",
    "queue_totals.messages_details.rate",
    "queue_totals.messages_ready",
    "queue_totals.messages_ready_details.rate",
    "queue_totals.messages_unacknowledged",
    "queue_totals.messages_unacknowledged_details.rate",
  ]
  name = ("overview." + (data.cluster_name
          if data.cluster_name else data.node.split ("@")[1]))
  aggregate (name, keys, data, stats)
  message_stats (name, "message_stats", data, stats)
  data = get_data (mp_conn, "api/queues")
  keys = [
    "consumer_utilisation",
    "consumers",
    "disk_reads",
    "disk_writes",
    "durable",
    "memory",
    "message_bytes",
    "message_bytes_persistent",
    "message_bytes_ram",
    "message_bytes_ready",
    "message_bytes_unacknowledged",
    "messages",
    "messages_persistent",
    "messages_ram",
    "messages_ready",
    "messages_ready_ram",
    "messages_unacknowledged",
    "messages_unacknowledged_ram",
  ]
  for d in data:
    name = "queues." + d.name
    if ignore_queues and pattern_match (name, ignore_queues):
      continue
    aggregate (name, keys, d, stats)
    key = "queues.{}.{}".format (name, "state")
    value = 1 if d["state"] == "running" else 0
    stats[key] = value
  data = get_data (mp_conn, "api/vhosts")
  keys = [
    "messages",
    "messages_details.rate",
    "messages_ready",
    "messages_ready_details.rate",
    "messages_unacknowledged",
    "messages_unacknowledged_details.rate",
    "recv_oct",
    "send_oct",
    "send_oct_details.rate",
  ]
  for d in data:
    name = "vhosts." + d.name.replace ("/", "_slash_")
    aggregate (name, keys, d, stats)
    message_stats (name, "messages_stats", d, stats)
  return stats



#! /usr/bin/env python

import collections, logging, logtool
from addict import Dict

LOG = logging.getLogger (__name__)

class FlattenDict (object):

  @logtool.log_call (log_args = False)
  def __init__ (self, data):
    self.data = data

  # @logtool.log_call (log_args = False)
  def flatten_list (self, l, parent_key = '', sep = '.'):
    items = []
    items.append ((parent_key + sep + "#" + "count", len (l)))
    for i in xrange (len (l)):
      new_key = parent_key + ("[%d]" % i if parent_key else "%d" % i)
      items.extend (self.flatten_item (l[i], new_key, sep = sep))
    return items

  # @logtool.log_call (log_args = False)
  def flatten_dict (self, d, parent_key = '', sep = '.'):
    items = []
    for k, v in d.items ():
      new_key = parent_key + sep + k if parent_key else k
      items.extend (self.flatten_item (v, new_key, sep = sep))
    return items

  # @logtool.log_call (log_args = False)
  def flatten_item (self, i, parent_key = '', sep = '.'):
    items = []
    if isinstance (i, collections.MutableMapping): # Dictionary
      items.extend (self.flatten_dict (i, parent_key, sep = sep))
    elif isinstance (i, collections.MutableSequence):
      items.extend (self.flatten_list (i, parent_key, sep = sep))
    else:
      items.append ((parent_key, i))
    return items

  @logtool.log_call (log_args = False, log_rc = False)
  def run (self, parent_key = '', sep = '.'):
    return Dict (self.flatten_item (self.data, parent_key, sep = sep))

@logtool.log_call (log_args = False)
def flatten_dict (data):
  return FlattenDict (data).run ()



#! /usr/bin/env python

from __future__ import absolute_import
import json, logging, logging.config, logtool, raven
import statsd, setproctitle, sys, time
from configobj import ConfigObj
from .mqreport import mqreport
from ._version import get_versions

LOG = logging.getLogger (__name__)
DEFAULT_PROCNAME = "coneyeye"
DEFAULT_CONF = "/etc/coneyeye/coneyeye.conf"
DEFAULT_LOGCONF = "/etc/coneyeye/logging.conf"

class InitialiseException (Exception):
  pass

class RuntimeException (Exception):
  pass

@logtool.log_call
def sentry_exception (conf, stats, e, message = None):
  sentry = raven.Client (conf["sentry_dsn"],
                         auto_log_stacks = True,
                         release = get_versions ()["version"])
  sentry_tags = {"component": "coneyeye"}
  logtool.log_fault (e, message = message)
  data = {
    "conf": conf,
    "stats": stats,
  }
  if message:
    data["message"] = message
  sentry.extra_context (data)
  try:
    exc_info = sys.exc_info()
    rc = sentry.captureException (exc_info, **sentry_tags)
    LOG.error ("Sentry filed: %s", rc)
  finally:
    del exc_info

@logtool.log_call (log_args = False)
def app_main (conf):
  try:
    if not conf.get ("sentry_dsn"):
      raise InitialiseException ("Missing configuration: sentry_dsn")
    mq_conn = {
      "url": conf.get ("rabbitmq_adminapi_url"),
      "user": conf.get ("rabbitmq_adminapi_user"),
      "passwd": conf.get ("rabbitmq_adminapi_passwd"),
    }
    statsd_conn = {
      "host": conf.get ("statsd_host", "127.0.0.1"),
      "port": conf.get ("statsd_port", 8125),
      "prefix": conf.get ("statsd_prefix", "rabbitmq"),
    }
    delay = int (conf.get ("delay", 5)) * 60
    # ignore_queues = ["*.celery@*", "*.celeryev.*", "*pidbox"]
    ignore_queues = [s.strip ()
                     for s in conf.get ("suppress_queues", "").split(",")]
  except Exception as e:
    logtool.log_fault (e)
    raise InitialiseException
  stats = None
  while True:
    try:
      with statsd.StatsClient (**statsd_conn).pipeline() as pipe:
        stats = mqreport (mq_conn, ignore_queues = ignore_queues)
        LOG.info (json.dumps (stats, indent = 2))
        for k, v in stats.items ():
          pipe.gauge (k, v)
        pipe.send ()
      LOG.info ("Sent...")
      time.sleep (delay)
    except KeyboardInterrupt:
      raise
    except Exception as e:
      sentry_exception (conf, stats, e)
      LOG.info ("Sleeping for a minute...")
      time.sleep (60)

@logtool.log_call
def main ():
  try:
    logging.config.fileConfig (DEFAULT_LOGCONF,
                               disable_existing_loggers = False)
    setproctitle.setproctitle (DEFAULT_PROCNAME)
    conf = ConfigObj (DEFAULT_CONF, interpolation = False)
    app_main (conf)
  except InitialiseException as e:
    print >> sys.stderr, "Configuration problem.  See logs for details."
    sys.exit (1)
  except KeyboardInterrupt:
    sys.exit (0)
  except RuntimeException:
    LOG.error ("Exiting...")
    sys.exit (1)
  except Exception as e:
    logtool.log_fault (e)
    LOG.error ("Exiting...")
    sys.exit (1)

if __name__ == "__main__":
  main ()



#! /usr/bin/env python


from setuptools import setup, find_packages
import glob, versioneer

setup (
    name = "coneyeye",
    version = versioneer.get_version (),
    description = "Daemon to map RabbitMQ stats into StatsD",
    long_description = file ("README.rst").read (),
    cmdclass = versioneer.get_cmdclass (),
    classifiers = [],
    keywords = "RabbitMQ, StatsD, daemon",
    author = "J C Lawrence",
    author_email = "claw@kanga.nu",
    url = "https://github.com/clearclaw/coneyeye",
    license = "GPL v3.0",
    packages = find_packages (exclude = ["tests"]),
    package_data = {"coneyeye": ["_cfgtool/coneyeye",
                                 "_cfgtool/*.templ",
                                 "_cfgtool/install",],
    },
    data_files = [
        ("/etc/cfgtool/module.d/", ["coneyeye/_cfgtool/coneyeye",]),
        ("/etc/coneyeye", glob.glob ("coneyeye/_cfgtool/*.templ")),
    ],
    zip_safe = False,
    install_requires = [line.strip ()
                        for line in file ("requirements.txt").readlines ()],
    entry_points = {
        "console_scripts": [
            "coneyeye = coneyeye.main:main",
        ],
    },
)
